/* 
 * xf.c --
 *
 *	A visual aid for selecting filenames
 *	from anywhere in the file system and outputting them
 *	on the standard output.  This file contains the
 *	main() and error-message routines.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Author: Gary Shea, UC Berkeley, Dept. of Chemistry
 */

#ifndef lint
static char rcsid[] = "$Header: xf.c,v 1.1 89/11/25 13:44:13 shea Exp $";
#endif /* not lint */

#include "xf.h"
#ifdef __hpux
#include <signal.h>
#endif
#define XF_APP_CLASS	"Xf"

#ifdef ERRFILE
#define ErrorFileName	"dt.err"
#endif

/* Local functions. */

static void	format( void ) ;
static void	Trap( int sig ) ;


/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	Initialize the X interface, the directory tree, build and
 *	display the initial directory box, handle any command-line-
 *	specified directories, then loop forever.
 *
 *----------------------------------------------------------------------
 */

main( int argc, char **argv )
{
    char	errLine[ ERRLINE_LEN ] ;
    char	*pLocName = argv[0] ;

    char	c ;
    int		getoptError = 0 ;

    extern int	optind ;
    extern char	*optarg ;

    extern Widget	BuildDirBox( DirTree *dirPtr ) ;


#ifdef ERRFILE
    /* Open the error file. */

    if ( ( pErrFp = fopen( ErrorFileName, "w" )) == (FILE *)0 )
    {
	sprintf( errLine, "Can't open error file <%s>", ErrorFileName ) ;
	FatalError( pLocName, errLine ) ;
    }
#else
    pErrFp = stderr ;
#endif

    /* Get X set up. */

    XtToolkitInitialize() ;
    appContext = XtCreateApplicationContext() ;

    if ( ( display = XtOpenDisplay( appContext,
	NULL, NULL, XF_APP_CLASS, NULL, 0, &argc, argv ) ) == NULL )
    {
	fprintf( stderr, "Unable to open display\n" );
	exit( 1 ) ;
    }

    /* Prepare to handle user-sent signals gracefully. */

    signal ( SIGINT, Trap ) ;
    signal ( SIGQUIT, Trap ) ;

    /* Yeah, it does seem a little pointless right now,
    *  but it's bound to grow...
    */

    while ( ( c = getopt ( argc, argv, "a" )) != EOF )
    {
	switch ( c )
	{
	case 'a' : ls_mode_a = TRUE ; break ;
	case '?' : ++getoptError ; break ;
	}
    }

    if ( getoptError ) format () ;

    topLevelW = XtAppCreateShell ( NULL, XF_APP_CLASS,
	applicationShellWidgetClass, display, NULL, 0 ) ;

#ifdef DEBUG
    XSynchronize ( display, True ) ;
#endif

    /* Initialize the directory tree to know where we are now,
    *  and that the current directory is where to root relative
    *  filenames.
    */

    DirTree_CWD() ;

#ifdef DEBUG
    fprintf( pErrFp, "Survived DirTree_CWD\n" );
#endif

    /* If the user specified any directories on the command line,
    *  handle them now.
    */

     if ( optind < argc )
     {
	int	ok = FALSE ;

	for ( ; optind < argc ; ++optind )
	{
	    DirTree		*dtPtr ;
    
	    /* Extend the directory tree if necessary... */
    
	    if ( ( dtPtr = DirTree_Find ( argv[ optind ] ) ) == NULL )
		continue ;
    
	    /* Display the directory box for the directory. */
	
	    if ( ! dtPtr->isDisplayed ) BuildDirBox ( dtPtr ) ;
	    ok = TRUE ;
	}

	if ( !ok ) exit( 1 ) ;
    }
    else
	BuildDirBox ( cwdPtr ) ;

    /* Loop it up. */

    XtAppMainLoop ( appContext ) ;
}


/*
 *----------------------------------------------------------------------
 *
 * OSFatalError, UserError, UserFatalError,
 *		FatalError, InternalError, Warning --
 *
 *	Global error-message routines.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	FatalError, InternalError, UserFatalError, OSFatalError
 *	exit the program.  All spew some output to the stderr.
 *
 *----------------------------------------------------------------------
 */

void
OSFatalError( char *routineName, char *message )
{
    char	errLine[ ERRLINE_LEN ] ;

    sprintf ( errLine,
	"fatal error: %s: %s\nos" ) ;
    perror ( errLine ) ;

    exit( 1 ) ;
}

void
UserError( char *message )
{
    fprintf( stderr, "%s\n", message ) ;
}

void
UserFatalError( char *message )
{
    UserError ( message ) ;
    exit ( 1 ) ;
}

void
FatalError( char *routineName, char *message )
{
    fprintf( stderr,
	"fatal error: %s: %s\n",
	routineName,
	message ) ;
    exit( 1 ) ;
}

void
InternalError( char *routineName, char *message )
{
    fprintf( stderr,
	"internal error: %s: %s\n",
	routineName,
	message ) ;
    exit( 1 ) ;
}

void
Warning( char *routineName, char *message )
{
    fprintf( stderr,
	"warning: %s: %s\n",
	routineName,
	message ) ;
}


/*
 *----------------------------------------------------------------------
 *
 * format --
 *
 *	Display the calling conventions (such as they are)
 *	of the xf program.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The program exits after output to stderr.
 *
 *  Arguments:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
format( void )
{
    char	errLine[ ERRLINE_LEN ] ;
    char	*pLocName = "format()" ;

#ifdef DEBUG
    fprintf ( pErrFp, "\n%s:\n", pLocName ) ;
#endif

    fprintf ( stderr, "Usage: xf [options] [name...]\n" ) ;
    fprintf ( stderr, "options: -a    Display files beginning with '.'\n'" ) ;
}


/*
 *----------------------------------------------------------------------
 *
 * Trap --
 *
 *	Trap the usual signals so that the application may escape
 *	gracefully.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The program is exited.
 *
 *  Arguments:
 *	arg - Description.
 *	arg - Description.
 *	arg - Description.
 *
 *----------------------------------------------------------------------
 */

static void
Trap( int sig )
{
    char	errLine[ ERRLINE_LEN ] ;
    char	*pLocName = "Trap ()" ;

# ifdef DEBUG
    fprintf ( pErrFp, "\n%s:\n", pLocName ) ;
# endif

    XtDestroyApplicationContext ( appContext ) ;
    exit ( 1 ) ;

}

